module Agda.Packaging.Types where

{-
-- External Library Imports
import qualified Distribution.InstalledPackageInfo
  as Cabal
    ( InstalledPackageInfo )

-------------------------------------------------------------------------------

type PackageDBName = FilePath
type PackageDB     = [Cabal.InstalledPackageInfo]

data NamedPackageDB
  =  NamedPackageDB
  { dbName :: PackageDBName
  , db     :: PackageDB }

type PackageDBStack = [NamedPackageDB]
-}
